@@ -1,34 +1,51 @@
 import base64
+import hashlib
 import json
 import logging
 import os
 import requests
-import subprocess
 import time
 import urllib.parse
 import uuid
+import warnings
+import webbrowser
+
 
 import io
-import PIL
+from distutils.util import strtobool
 from PIL import Image
+from packaging import version
+import selenium
+
+if version.parse(selenium.__version__) < version.parse('4.0.0'):
+    old_selenium = True
+else:
+    old_selenium = False
 
 from selenium.webdriver.common.action_chains import ActionChains
 
 from selenium import webdriver
 
-import requests.packages.urllib3
 requests.packages.urllib3.disable_warnings()
 
+
+log = logging.getLogger(__name__)
+
 class TestAiDriver():
-    def __init__(self, driver, api_key, server_url=None, test_case_name=None, debug=False, train=False, use_classifier_during_creation=True):
+    def __init__(self, driver, api_key, test_case_name=None, debug=False, use_classifier_during_creation=True, train=False, server_url=None):
         self.debug = debug
         self.train = train
         self.driver = driver
         self.api_key = api_key
         self.run_id = str(uuid.uuid1())
-        self.test_case_creation_mode = test_case_name is not None
         self.last_test_case_screenshot_uuid = None
+        try:
+            self.test_case_creation_mode = strtobool(os.environ.get('TESTAI_INTERACTIVE', '0')) == 1
+        except Exception:
+            self.test_case_creation_mode = False
         if self.test_case_creation_mode:
+            if test_case_name is None:
+                test_case_name = os.path.basename(__file__).split('.py')[0]
             self.test_case_uuid = test_case_name
             self.use_classifier_during_creation = use_classifier_during_creation
         if server_url is None:
@@ -41,6 +58,10 @@ class TestAiDriver():
         width, height = im.size
         self.multiplier = 1.0 * width / window_size['width']
         self_attrs = dir(self)
+        # Disable warnings
+        requests.packages.urllib3.disable_warnings()
+        warnings.filterwarnings("ignore", category=DeprecationWarning)
+
         for a_name in dir(driver):
             if a_name in self_attrs:
                 continue
@@ -70,7 +91,6 @@ class TestAiDriver():
                     setattr(self, a_name, v)
             except Exception as err:
                 continue
-                # print(err)
 
 
     def implicitly_wait(self, wait_time):
@@ -100,10 +120,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element(by=by, value=value)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -140,10 +162,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_accessibility_id(accessibility_id)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -180,10 +204,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_class_name(name)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -222,10 +248,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_css_selector(css_selector)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -264,10 +292,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_id(id_)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -306,10 +336,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_link_text(link_text)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -349,10 +381,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_name(name)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -391,10 +425,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_partial_link_text(link_text)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -433,10 +469,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_tag_name(name)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -475,10 +513,12 @@ class TestAiDriver():
         # Run the standard selector
         try:
             driver_element = self.driver.find_element_by_xpath(xpath)
-            if driver_element and self.train == True:
-                classified_element, key, msg = self._classify(element_name)
+            if driver_element:
+                key = self._upload_screenshot_if_necessary(element_name)
                 self._update_elem(driver_element, key, element_name)
             return driver_element
+        except NoElementFoundException as e:
+            log.exception(e)
         except Exception as err:
             # If this happens, then error during the driver call
             classified_element, key, msg = self._classify(element_name)
@@ -508,12 +548,12 @@ class TestAiDriver():
                 element = driver.find_element_by_xpath('//div/td[1]')
         """
         el, key, msg = self._classify(element_name)
+
         if el is None:
             raise Exception(msg)
         return el
 
     def _update_elem(self, elem, key, element_name, train_if_necessary=True):
-        # print('update_elem', type(elem.rect['x']))
         data = {
             'key': key,
             'api_key': self.api_key,
@@ -545,7 +585,8 @@ class TestAiDriver():
                 return element, self.last_test_case_screenshot_uuid, msg
             else:
                 label_url = self.url + '/test_case/label/' + urllib.parse.quote(self.test_case_uuid)
-                print('Waiting for bounding box of element {} to be drawn in the UI {}'.format(element_name, label_url))
+                log.info('Waiting for bounding box of element {} to be drawn in the UI: \n\t{}'.format(element_name, label_url))
+                webbrowser.open(label_url)
                 while True:
                     element_box = self._test_case_get_box(element_name)
                     if element_box is not None:
@@ -561,6 +602,15 @@ class TestAiDriver():
             # Call service
             ## Get screenshot & page source
             screenshotBase64 = self.driver.get_screenshot_as_base64()
+            key = self.get_screenshot_hash(screenshotBase64)
+            resp_data = self._check_screenshot_exists(key, element_name)
+            if resp_data['success'] and 'box' in resp_data:
+                if self.debug:
+                    print(f'Found cached box in action info for {element_name} using that')
+                real_elem = self._match_bounding_box_to_selenium_element(resp_data['box'], multiplier=self.multiplier)
+                element = testai_elem(real_elem.parent, real_elem, resp_data['box'], self.driver,
+                                      self.multiplier)
+                return element, key, msg
 
             # ZZZ
             with open('/tmp/scnshot.png', 'wb') as outFile:
@@ -583,7 +633,7 @@ class TestAiDriver():
                 response = json.loads(r.text)
                 run_key = response['key']
                 if response.get('success', False):
-                    print('successful classification of element_name: %s' % element_name)
+                    log.info('successful classification of element_name: %s' % element_name)
                     pred_elem = response['elem']
                     real_elem = self._match_bounding_box_to_selenium_element(pred_elem, multiplier=self.multiplier)
                     element = testai_elem(real_elem.parent, real_elem, pred_elem, self.driver,
@@ -594,6 +644,56 @@ class TestAiDriver():
                 logging.exception('exception during classification')
             return element, run_key, msg
 
+    def get_screenshot_hash(self, b64img):
+        msg = base64.b64decode(b64img)
+        buf = io.BytesIO(msg)
+        img = Image.open(buf)
+        w, h = img.size
+        return hashlib.md5(img.crop((0, 75, w - 50, h - 75)).tobytes()).hexdigest()
+
+    def _check_screenshot_exists(self, key, element_name):
+        data = {'api_key': self.api_key, 'screenshot_uuid': key, 'label': element_name}
+        check_screenshot_url = self.url + '/check_screenshot_exists'
+        start = time.time()
+        r = requests.post(check_screenshot_url, json=data, verify=False)
+        end = time.time()
+        if self.debug:
+            print(f'Cached bounding box request time: {end - start}')
+
+        if r.status_code != 200:
+            raise Exception('Error checking cached screenshot from remote')
+        else:
+            response = json.loads(r.text)
+            return response
+
+    def _upload_screenshot_if_necessary(self, element_name):
+        screenshotBase64 = self.driver.get_screenshot_as_base64()
+        key = self.get_screenshot_hash(screenshotBase64)
+        # Check results
+        try:
+            response = self._check_screenshot_exists(key, element_name)
+            if response['success'] == True:
+                if self.debug:
+                    print(f'Screenshot {key} already exists on remote')
+                return key
+            else:
+                if self.debug:
+                    print(f'Screenshot {key} does not exist on remote, uploading it')
+                data = {'api_key': self.api_key, 'screenshot_uuid': key, 'screenshot': screenshotBase64, 'label': element_name}
+                upload_screenshot_url = self.url + '/upload_screenshot'
+                start = time.time()
+                r = requests.post(upload_screenshot_url, json=data, verify=False)
+                end = time.time()
+                if self.debug:
+                    print(f'Upload screenshot request time: {end - start}')
+                if r.status_code != 200:
+                    log.error('Error uploading screenshot to remote')
+                else:
+                    pass
+                return key
+        except Exception:
+            log.exception('Error checking cached screenshot / uploading it from remote')
+
 
     def _test_case_get_box(self, label):
         """
@@ -642,7 +742,6 @@ class TestAiDriver():
         # Adapt box to local coordinates
         new_box = {'x': bounding_box['x'] / multiplier, 'y': bounding_box['y'] / multiplier,
                    'width': bounding_box['width'] / multiplier, 'height': bounding_box['height'] / multiplier}
-
         # Get all elements
         elements = self.driver.find_elements_by_xpath("//*")
         # Compute IOU
@@ -661,10 +760,10 @@ class TestAiDriver():
         composite = list(filter(lambda x: self._center_hit(new_box, x[1].rect), composite))
 
         if len(composite) == 0:
-            raise Exception('Could not find any web element under the center of the bounding box')
+            raise NoElementFoundException('Could not find any web element under the center of the bounding box')
         else:
             for score, element in composite:
-                if element.tag_name == 'input' or element.tag_name == 'button':
+                if (element.tag_name == 'input' or element.tag_name == 'button') and score > composite[0][0] * 0.9:
                     return element
             return composite[0][1]
 
@@ -696,7 +795,10 @@ class TestAiDriver():
 class testai_elem(webdriver.remote.webelement.WebElement):
     def __init__(self, parent, source_elem, elem, driver, multiplier=1.0):
         # We need to also pass the _w3c flag otherwise the get_attribute for thing like html or text is messed up
-        super(testai_elem, self).__init__(source_elem.parent, source_elem._id)
+        if old_selenium:
+            super(testai_elem, self).__init__(source_elem.parent, source_elem._id, w3c=source_elem._w3c)
+        else:
+            super(testai_elem, self).__init__(source_elem.parent, source_elem._id)
         self.driver = driver
         self._text = elem.get('text', '')
         self._size = {'width': elem.get('width', 0)/multiplier, 'height': elem.get('height', 0)/multiplier}
@@ -706,7 +808,7 @@ class testai_elem(webdriver.remote.webelement.WebElement):
         self.rect.update(self._size)
         self.rect.update(self._location)
         self._tag_name = elem.get('class', '')
-        self._cx = elem.get('x', 0)/multiplier  + elem.get('width', 0) / multiplier / 2
+        self._cx = elem.get('x', 0)/multiplier + elem.get('width', 0) / multiplier / 2
         self._cy = elem.get('y', 0)/multiplier + elem.get('height', 0) /multiplier / 2
 
     @property
@@ -747,3 +849,6 @@ class testai_elem(webdriver.remote.webelement.WebElement):
         return True
     def is_displayed(self):
         return True
+
+class NoElementFoundException(Exception):
+    pass
\ No newline at end of file
